# -*- coding: utf-8 -*-

import numpy as np
import copy
from circadapt.error import ModelNotStable


class ModelAdapt():
    """Model class to be inherited to give model adaptation protocol."""

    def __init__(self):
        self._wall_loc = [
            'Model.Peri.La.wLa.',
            'Model.Peri.Ra.wRa.',
            'Model.Peri.TriSeg.wLv.',
            'Model.Peri.TriSeg.wSv.',
            'Model.Peri.TriSeg.wRv.',
            ]
        self._patch_loc = [
            'Model.Peri.La.wLa.pLa1.',
            'Model.Peri.Ra.wRa.pRa1.',
            'Model.Peri.TriSeg.wLv.pLv1.',
            'Model.Peri.TriSeg.wSv.pSv1.',
            'Model.Peri.TriSeg.wRv.pRv1.',
            ]

        self._all_senses = ['SfEcmMax',
                            'SfActMax',
                            'SfPasAct',
                            'LsPasAct',
                            ]
        self._all_senses_norm = ['SfPasMaxT',
                                 'FacSfActT',
                                 'SfPasActT',
                                 'LsPasActT',
                                 ]
        self._all_actors = ['VWall',
                            'AmRef',
                            'SfPas',
                            ]

        self._artven_cavities = [
            'Model.SyArt.',
            'Model.SyVen.',
            'Model.PuArt.',
            'Model.PuVen.',
            ]
        self._artven_names = [
            'SyArt',
            'SyVen',
            'PuArt',
            'PuVen',
            ]

        self._artven_actors = [
            'A_wall',
            'A0',
            'p0',
            ]

        self._valve_names = [
            'Model.Peri.SyVenRa.',
            'Model.Peri.RaRv.',
            'Model.Peri.RvPuArt.',
            'Model.Peri.PuVenLa.',
            'Model.Peri.LaLv.',
            'Model.Peri.LvSyArt.',
            ]

        self._valve_actors = [
            'A_open',
            ]

        self._bags = ['Model.Peri.']

        self._adapt_options = {
            'rest': {
                'tCycle': 0.85,
                'q0': 8.5e-5,
                'ventricles': {
                    },
                'General': {},
                },
            'exercise': {
                'tCycle': 0.85 / 2,
                'q0': 8.5e-5 * 3,
                'ventricles': {
                    },
                'General': {},
                },
            'protocol': {
                'n_beats': 50,
                'n_cycles': 3,
                'adapt_Wall': True,
                'adapt_Bag': True,
                'adapt_ArtVen': True,
                'adapt_Valve': True,
                },
            }

    _adapt_patch_object = 'Patch2022'

    def get_adapt_options(self):
        """Get default adapt options."""
        options = copy.deepcopy(self._adapt_options)
        return options

    def _update_options(self, _options):
        # first update options
        options = self.get_adapt_options()
        options.update(_options)

        # now check for exercise ventricular parameters
        # set key in rest if not exist to current value assuming it is in rest
        for key in options['exercise']['ventricles']:
            if key not in options['rest']['ventricles']:
                options['rest']['ventricles'][key] = \
                    self[self._adapt_patch_object][key][2:]

        for key in options['exercise']['General']:
            if key not in options['rest']['General']:
                options['rest']['General'][key] = \
                    self['General'][key]
        return options

    def adapt(self,
              options: dict = {},
              output: bool = False,
              ) -> None:
        """
        Run adaptation protocol.

        The adaptation protocol runs n_cycles cycles. Each cycle has two
        phases, namely rest adaptation and exercise adaptation. First, in
        exercise, the Patches and vessel wall volumes are adapted to load. In
        rest, vessels are adapted to flow.

        Parameters
        ----------
        options : dictionary, optional
            Options for the protocol. To set options, use the function
            get_adapt_options(), which is used by default. The default is {}.
        output : bool, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        senses_results : TYPE
            Only when output=True.
        senses_norm : TYPE
            Only when output=True.
        actor_results : TYPE
            Only when output=True.
        actor_vwall : TYPE
            Only when output=True.

        """
        options = self._update_options(options)

        # get options
        n_beats = options['protocol']['n_beats']
        n_cycles = options['protocol']['n_cycles']

        # allocate
        if output:
            (senses_results, senses_norm,
             actor_results, actor_vwall, actor_valve) = \
                self._allocate_output(n_beats * n_cycles * 2)

        # run adaptation
        for i in range(n_beats * n_cycles * 2):
            if (i // n_beats) % 2 == 0:
                self._run_simulation(options['exercise'])
                self.adapt_exercise(options)
            else:
                self._run_simulation(options['rest'])
                self.adapt_rest(options)

            if output:
                self._adapt_store_data(
                    i,
                    senses_results,
                    senses_norm,
                    actor_results,
                    actor_vwall,
                    actor_valve,
                    )

        # Bring back to rest
        self._run_simulation(options['rest'])

        if output:
            return (senses_results, senses_norm,
                    actor_results, actor_vwall, actor_valve)

    def _allocate_output(self, nBeats):
        senses_results = np.ndarray((nBeats,
                                     len(self._patch_loc),
                                     len(self._all_senses)))
        senses_norm = np.ndarray((nBeats,
                                  len(self._patch_loc),
                                  len(self._all_senses)))
        actor_results = np.ndarray((nBeats,
                                    len(self._patch_loc),
                                    len(self._all_actors)))
        actor_vwall = np.ndarray((nBeats,
                                  len(self._artven_names),
                                  len(self._artven_actors)))
        actor_valve = np.ndarray((nBeats,
                                  len(self._valve_names),
                                  len(self._valve_actors)))
        return (senses_results, senses_norm,
                actor_results, actor_vwall, actor_valve)

    def adapt_exercise(self, options):
        """Trigger all excercise adaptation functions."""
        if options['protocol']['adapt_Wall']:
            for wall in self._wall_loc:
                self.trigger(wall+'adaptExercise')

        if options['protocol']['adapt_ArtVen']:
            for avc in self._artven_cavities:
                self.trigger(avc+'adaptExercise')

        if options['protocol']['adapt_Bag']:
            for bag in self._bags:
                self.trigger(bag+'adaptExercise')

    def adapt_rest(self, options):
        """Trigger all rest adaptation functions."""
        if options['protocol']['adapt_ArtVen']:
            for avc in self._artven_cavities:
                self.trigger(avc+'adaptRest')

        if options['protocol']['adapt_Valve']:
            for valve in self._valve_names:
                self.trigger(valve+'adaptRest')

    def _run_simulation(self, sub_options):
        self._set_parameters(sub_options)
        self.run(run_stable=True)

        # stability essential for adaptation
        if not self.get('Model.isStable'):
            raise ModelNotStable()

    def _set_parameters(self, sub_options):
        self.set('Model.tCycle', sub_options['tCycle'])
        self.set('Model.PFC.q0', sub_options['q0'])

        # ventricles
        for key, value in sub_options['ventricles'].items():
            self[self._adapt_patch_object][key][2:] = value

        for key, value in sub_options['General'].items():
            self['General'][key] = value

    def _adapt_store_data(
            self,
            i,
            senses_results,
            senses_norm,
            actor_results,
            actor_vwall,
            actor_valve,
            ):

        # patch
        for i_wall, wall in enumerate(self._patch_loc):
            for i_sens, sens in enumerate(self._all_senses):
                senses_results[i, i_wall, i_sens] = self.get(wall+sens,
                                                             dtype=float)
                s = wall+self._all_senses_norm[i_sens]
                senses_norm[i, i_wall, i_sens] = self.get(s, dtype=float)
                if self._all_senses_norm[i_sens] == 'FacSfActT':
                    senses_norm[i, i_wall, i_sens] *= self.get(wall+'SfAct',
                                                               dtype=float)

            for i_act, act in enumerate(self._all_actors):
                actor_results[i, i_wall, i_act] = self.get(wall+act,
                                                           dtype=float)

        # artven
        for i_wall, loc in enumerate(self._artven_cavities):
            for i_act, act in enumerate(self._artven_actors):
                actor_vwall[i, i_wall, i_act] = self.get(loc+act, dtype=float)

        # valve
        for i_valve, loc in enumerate(self._valve_names):
            for i_act, act in enumerate(self._valve_actors):
                actor_valve[i, i_valve, i_act] = self.get(loc+act, dtype=float)

    def _set_matrix(self):
        mean_dlogpar_dlogsens = self.calculate_matrix()
        for patch in [
                'Model.Peri.La.wLa.pLa1',
                'Model.Peri.Ra.wRa.pRa1',
                'Model.Peri.TriSeg.wLv.pLv1',
                'Model.Peri.TriSeg.wSv.pSv1',
                'Model.Peri.TriSeg.wRv.pRv1',
                ]:
            for i in range(3):
                for j in range(4):
                    self.set(patch+f'.transmat{i}{j}',
                              mean_dlogpar_dlogsens[i, j]*1e-3)


    def calculate_matrix(self):
        state = self.model_export()

        self.load_reference()

        # CA.load('../test_noAdapt_fit10/ref000.npy')
        self.run(run_stable=True)


        reference_y = np.array([
            self['Patch2022']['SfEcmMax'],
            self['Patch2022']['SfActMax'],
            self['Patch2022']['SfPasAct'],
            self['Patch2022']['LsPasAct'],
            ]).T

        n_patch = 5
        n_par = 3

        par = ['VWall', 'AmRef', 'SfPas']

        model_parameters = np.ndarray((n_patch, n_par, 2))
        sensing_values = np.ndarray((n_patch, n_par, 2, 4))

        for i_patch in range(n_patch):
            for i_par in range(n_par):
                for i_sign, sign in enumerate([-1, 1]):
                    self.load_reference()

                    self['Patch2022'][par[i_par]][i_patch] *= 1+0.05*sign
                    model_parameters[i_patch, i_par, i_sign] = self['Patch2022'][par[i_par]][i_patch]
                    self.run(run_stable=True)


                    sensing_values[i_patch, i_par, i_sign, :] = np.array([
                        self['Patch2022']['SfEcmMax'][i_par],
                        self['Patch2022']['SfActMax'][i_par]/self['Patch2022']['SfAct'][i_par],
                        self['Patch2022']['SfPasAct'][i_par],
                        self['Patch2022']['LsPasAct'][i_par],
                        ])

        sens = reference_y.reshape(5, 1, 1, 4) / sensing_values
        log_sens = np.log(sens)
        # log_sens = np.log(1/sensing_values)
        # diff_log_sens = np.diff(log_sens, axis=2)[:, :, 0, :]
        diff_log_sens = log_sens[:, :, 1, :] - log_sens[:, :, 0, :]

        log_parameters = np.log(model_parameters)
        diff_parameters = log_parameters[:, :, 1] - log_parameters[:, :, 0]

        diff_log_sens_Lv = diff_parameters[2].reshape((-1, 1)) / diff_log_sens[2, :, :]

        mean_dlogpar_dlogsens = np.mean(
            diff_parameters.reshape((*diff_parameters.shape, 1)) / diff_log_sens,
            axis=0)

        self.model_import(state)

        return mean_dlogpar_dlogsens

