"""Cavity objects."""
from circadapt.components import Component


class Cavity(Component):
    """Basic Cavity object."""

    parameters = [
        ]

    signals = [
        'A',
        'p',
        'p_trans',
        'q',
        'V',
        'VDot'
        ]


class Bag(Component):
    """
    Bag is designed to simulate the pericardium

    Parameters
    ----------
    k: double
        Nonlinear exponent
    p_ref: double
        Reference pressure for which hold p(V_ref) = p_ref
    V_ref: double
        Reference volume for which hold p(V_ref) = p_ref

    Signals
    -------
    V: array
        Volume enclosed by the Bag
    p: array
        Pressure inside the Bag
    """

    parameters = [
        'k',
        'p_ref',
        'V_ref',
        ]

    signals = [
        'V',
        'p',
        ]


class Capacitor(Cavity):
    """
    Bag is designed to simulate the pericardium

    Parameters
    ----------
    Vo: double
        Reference volume for which hold p=0
    C: double
        Linear capacitance of cavity

    Signals
    -------
    V: array
        Volume enclosed by the Bag
    p: array
        Pressure inside the Bag
    """

    parameters = [
        'V0',
        'C',
        ]

    signals = [
        'V',
        ]


class Tube0D(Component):
    """
    Tube0D is designed to simulate pressure-volume relations of vessels.

    Parameters
    ----------
    l: double
        Length of the tube
    A_wall: double
        Cross-sectional area of the wall
    k: double
        Stiffness coefficient of the wall
    p0: double
        Pressure at A=A0
    A0: double
        Cross-sectional cavity area for p=p0

    Signals
    -------
    V: array
        Volume
    p: array
        Pressure
    """

    parameters = [
        'l',
        'A_wall',
        'k',
        'p0',
        'A0',
        ]

    signals = [
        'V',
        'p',
        ]


class Chamber2022(Component):
    """
    Chamber2022 object.

    Parameters
    ----------
    buckling: bool
        Buckling function. If True, wall tension cannot be below zero.

    Signals
    -------
    V: array
        Volume
    p: array
        Pressure
    """

    parameters = [
        'buckling',
        ]

    signals = [
        'V',
        'p',
        'V',
        'VDot',
        'A',
        'Y',
        'q',
        ]

    def add_object(self, loc):
        super().add_object(loc)
        self._model._check_component_in_export_list_and_add_object(
            'Wall2022',
            loc+'.w'+loc[-2:],
            )


class TriSeg2022(Component):
    """
    TriSeg2022 is designed to represent ventricles including interaction.


    Parameters
    ----------
    tau: float [-]
        Constant to limit change in estimation of dV and dY

    Signals
    -------
    V: array
        Volume of truncated septal wall
    V0: array
        Volume of truncated septal wall before finding mechanical equilibrium
    VDot: array
        Time-derivative of V, used to find the estimation V0
    Y: array
        Distance of wall-junction to center
    Y0: array
        Distance of wall-junction to center before finding mechanical
        equilibrium
    YDot: array
        Time-derivative of Y, used to find the estimation Y0
    """

    parameters = [
        'tau',
        ]

    signals = [
        'V',
        'Y',
        'V0',
        'Y0',
        'VDot',
        'YDot',
        'nLoop',
        'Fp',
        'Fv',
        ]

    def add_object(self, loc):
        super().add_object(loc)
        self._model._check_component_in_export_list_and_add_object('Cavity',
                                                                   loc+'.cLv')
        self._model._check_component_in_export_list_and_add_object('Cavity',
                                                                   loc+'.cRv')


        self._model._check_component_in_export_list_and_add_object('Wall2022',
                                                                   loc+'.wLv')
        self._model._check_component_in_export_list_and_add_object('Wall2022',
                                                                   loc+'.wSv')
        self._model._check_component_in_export_list_and_add_object('Wall2022',
                                                                   loc+'.wRv')
