"""Patch objects."""
from circadapt.components import Component


class Patch2022(Component):
    """
    Patch2022 is based on Patch in Walmsley 2015.

    Parameters
    ==========
    AmRef: float
        Reference wall area at :math:`l_{s} = l_{s,ref}`.
    VWall: float
        Wall volume
    vMax: float
        Maximum shortening velocity
    l_se: float
        Length of the series elastic element, i.e. :math:`l_{s} -l_{si}` for
        which stress is zero.
    LsRef: float
        Reference sarcomere length for which at
        :math:`A_m (l_{s,ref}) = A_{m,ref}`.
    Ls0Pas: float
        Zero passive stress length
    dLsPas: float
        -
    SfPas: float
        Linear ECM stress coefficient
    k1: float
        Nonlinear exponent ECM stress component
    TR: float
        Contraction time constant
    TD: float
        Relaxation time constant
    TimeAct: float
        Relative contraction duration
    SfAct: float
        Linear active stress component
    dT: float
        Activation delay relative to intrinsic activation
    CRest: float
        Rest contractility
    Lsi0Act: float
        Reference length for zero-active-stress
    LDAD: float
        length dependend activation duration
    ADO: float
        activation duration offset
    LDCC: float
        length dependend contractility coefficient
    SfPasMaxT: float
        Maximum ecm stress (adaptation sens variable)
    SfPasActT: float
        Active weighted passive stress (adaptation sens variable)
    FacSfActT: float
        Active stress (adaptation sens variable)
    LsPasActT: float
        Weighted sarcomere length average (adaptation sens variable)
    adapt_gamma: bool
        Adaptation constant

    Signals
    =======
    Signals are arrays. Each point in the array represents a point in time with
    step-size controlled by the solver.

    Ls: array
        Sarcomere length
    Lsi: array
        State variable: Intrinsic sarcomere length
    LsiDot: array
        State variable: Intrinsic sarcomere length time-derivative
    C: array
        State variable: contraction curve
    CDot: array
        State variable: contraction time-derivative
    Am: array
        Patch mid-wall area
    Am0: array
        Patch mid-wall zero-stress area
    Ef: array
        Natural strain
    T: array
        Mid-wall tension
    DADT: array
        Area-tension derivative
    Sf: array
        Total fibre stress at mid-wall
    SfPasT: array
        Total passive stress at mid-wall
    SfEcm: array
        Total ECM stress at mid-wall
    DSfDEf: array
        Total stiffness coefficient
    DSfPasDEf: array
        Total passive stiffness coefficient
    SfEcmMax: array
        Adaptation: Maximum ECM stress
    SfActMax: array
        Adaptation: maximum active stress
    SfPasAct: array
        Adaptation: active-weighted passive stress
    LsPasAct: array
        Adaptation: active-weigthed sarcomere length

    """

    parameters = [
        'AmRef',
        'VWall',
        'vMax',
        'l_se',
        'LsRef',
        'Ls0Pas',
        'dLsPas',
        'SfPas',
        'TR',
        'TD',
        'TimeAct',
        'SfAct',
        'k1',
        'dT',
        'CRest',
        'Lsi0Act',
        'LDAD',
        'ADO',
        'LDCC',
        'SfPasMaxT',
        'SfPasActT',
        'FacSfActT',
        'LsPasActT',
        'adapt_gamma',
        ]

    signals = [
        'Ls',
        'Lsi',
        'LsiDot',
        'C',
        'CDot',
        'Am',
        'Am0',
        'Ef',
        'T',
        'DADT',
        'Sf',
        'SfPasT',
        'SfEcm',
        'DSfDEf',
        'DSfPasDEf',
        'SfEcmMax',
        'SfActMax',
        'SfPasAct',
        'LsPasAct'
        ]

    def __init__(self, model, objects=None, wall_locs=None):
        """
        Patch2022 is based on Patch in Walmsley 2015.

        Number of patches and their names may vary and are automatically
        detected from the model object.

        Parameters
        ----------
        model : circadapt.Circadapt
            Model object.
        objects : list, optional
            List of objects in c++ this object refers to. If not given,
            the list is build based on wall_locs.
        wall_locs : list, optional
            List of wall locations in c++. This list is used to generate the
            objects.

        """
        self._model = model
        self._wall_locs = wall_locs
        if objects is None and wall_locs is None:
            objects = []
        elif objects is None:
            objects = self.get_all_patches()
        super().__init__(model, objects)

    def build(self):
        """
        (re)build patch locations based on wall locations.

        There is no compatibility with different types of patches used in the
        model. If you use different type of patches, please use the direct way
        of interacting with the c++ object, i.e. set() and get().
        """
        if self._wall_locs is not None:
            self.objects = self.get_all_patches()
        super().build()

    def get_all_patches(self):
        """Get all patches of the walls included in this pointer object."""
        all_patches = []
        for i_wall, wall in enumerate(self._wall_locs):
            n_patch = self._model.get(wall + '.n_patch', int)
            for i_patch in range(n_patch):
                all_patches.append(wall+'.p'+wall[-2:]+str(i_patch+1))
        return all_patches
