"""Wall objects."""
from circadapt.components import Component


def rebuild_patch2022(model):
    """Rebuilds patches of the model, e.g. for when n_patch changed."""
    model['Patch2022'].build()


class Wall2022(Component):
    """
    Wall2022 object.

    Parameters
    ==========
    AmDead: float
        Non-conctractile wall area.
    n_patch: int
        Number of patches in the wall. On increase, the last patch will be
        copied. Wall volume and area are distributed equally over the patches.
        On decrease, the last patches in the list will be removed.

    Signals
    =======
    Signals are arrays. Each point in the array represents a point in time with
    step-size controlled by the solver.

    Am: array
        Wall area
    Am0: array
        Linearized stress-free wall area
    Cm: array
        Wall curvature
    DADT: array
        Stress-tension derivative
    p_trans: array
        Transmural pressure
    T: array
        Tension
    Vm: array
        Mid-wall volume
    VWall: double
        Wall volume
    Z: array
        Z
    """

    parameters = [
        'AmDead',
        'n_patch',
        ]
    parameter_on_set = {
        # default patch2022, overwrite when other patches are used
        'n_patch': rebuild_patch2022
        }

    signals = [
        'A',
        'Am',
        'Am0',
        'Cm',
        'DADT',
        'p_trans',
        'T',
        'Vm',
        'VWall',
        'Z',
        ]


class PreAfterloadExperiment(Component):
    """
    PreAfterloadExperiment object.

    Parameters
    ==========
    AmDead: float
        Non-conctractile wall area.
    AmRef_Afterload: float
        Wall area at :math:`T_{m} = Tafterload`.
    n_patch: int
        Number of patches. On change, the last patch is splitted or last
        patches are removed.
    n_iter: int
        Number of iterations to calculate mechanical stability.
    T_afterload: float
        Maximum wall tension. When tension is reached, wall area shortens.

    Signals
    =======
    Signals are arrays. Each point in the array represents a point in time with
    step-size controlled by the solver.

    Am: array
        Wall area
    Am0: array
        Linearized stress-free wall area
    Cm: array
        Wall curvature
    DADT: array
        Stress-tension derivative
    T: array
        Tension
    p_trans: array
        Transmural pressure
    """
    parameters = [
        'AmDead',
        'AmRef_Afterload',
        'n_patch',
        'n_iter',
        'T_afterload',
        ]
    parameter_on_set = {
        'n_patch': rebuild_patch2022,
        }
    signals = [
        'A',
        'Am',
        'Am0',
        'Cm',
        'DADT',
        'p_trans',
        'T',
        'Vm',
        'VWall',
        'Z',
        ]
