"""
VanOsta2023 model.

"""

import circadapt
from circadapt.model import Model

class VanOsta2023(Model):
    def __init__(self, solver='SolverCA'):
        Model.__init__(self, solver)

    def build(self):
        # Circulation
        self.add_smart_component('ArtVen', build='SystemicCirculation')
        self.add_smart_component('ArtVen', build='PulmonaryCirculation')
        self.add_smart_component('Heart', patch_type='Patch2022')
        self.add_smart_component('Timings')
        self.add_smart_component('PressureFlowControl')

    def set_reference(self):
        self['Chamber2022']['buckling'] = False
        self['Valve2022']['soft_closure'] = True

        self['Solver']['dT'] = 0.0005
        self['Solver']['dTexport'] = self['Solver']['dT']

        self.set('Model.tCycle', 0.85)

        self.set('Model.PFC.fac',  0.5)
        self.set('Model.PFC.epsilon', 0.1)
        self.set('Model.PFC.fac_pfc', 1)
        self.set('Model.PFC.stable_thresh', 0.001)

        # Set parameters Circulation
        self['Tube0D']['A0'] = [
            0.00049833,
            0.00050034,
            0.0004684 ,
            0.00051849,
            ]
        self['Tube0D']['A_wall'] = [
            1.14408188e-04,
            4.81082964e-05,
            5.19472921e-05,
            5.79933632e-05,
            ]
        self['Tube0D']['k'] = [ 8., 10.,  8., 10.]
        self['Tube0D']['l'] = [0.4, 0.4, 0.2, 0.2]
        self['Tube0D']['p0'] = [
            12161.21763497,
            112.09788908,
            1899.95009955,
            572.39790382,
            ]

        self['ArtVen']['p0'] = [6395.43973044,  500.]
        self['ArtVen']['q0'] = [4.5e-05, 4.5e-05]
        self['ArtVen']['k'] = [1, 2]

        # Set volume state variables
        # not needed to reset, only if initial model crashes
        if True:
            self.set('Model.SyArt.V', 2e-4)
            self.set('Model.PuArt.V', 1e-4)
            self.set('Model.SyVen.V', 4e-4)
            self.set('Model.PuVen.V', 2e-4)

            self.set('Model.Peri.TriSeg.V', 44e-6)
            self.set('Model.Peri.TriSeg.Y', 34.6e-3)

            self.set('Model.Peri.TriSeg.cLv.V',  150e-6)
            self.set('Model.Peri.TriSeg.cRv.V',  100e-6)
            self.set('Model.Peri.La.V',  100e-6)
            self.set('Model.Peri.Ra.V',  50e-6)

        # all wall parameters
        self['Patch2022']['l_se'] = 0.04
        self['Patch2022']['LsRef'] = 2.0
        self['Patch2022']['Ls0Pas'] = 1.8
        self['Patch2022']['dLsPas'] = 0.6
        self['Patch2022']['k1'] = 10
        self['Patch2022']['dT'] = 0
        self['Patch2022']['CRest'] = 0
        self['Patch2022']['Lsi0Act'] = 1.51
        self['Patch2022']['LDAD'] = 1.057
        self['Patch2022']['ADO'] = 0.65
        self['Patch2022']['LDCC'] = 4.
        self['Patch2022']['vMax'] = 7.
        self['Patch2022']['TR'] = 0.25
        self['Patch2022']['TD'] = 0.25

        # wall specific
        self['Patch2022']['AmRef'] = [
            0.00706159,
            0.00603949,
            0.0093191 ,
            0.00507053,
            0.01313792,
            ]
        self['Patch2022']['VWall'] = [
            1.86518494e-05,
            6.12511827e-06,
            9.51121820e-05,
            3.74835076e-05,
            4.97478018e-05,
            ]
        self['Patch2022']['SfPas'] = [
            51591.64007141,
            52277.48834349,
            22370.46712926,
            22051.80366169,
            23278.22158555,
            ]
        self['Patch2022']['SfAct'] = [84e3, 84e3, 120e3, 120e3, 120e3]

        # Valves
        self['Valve2022']['adaptation_A_open_fac'] = [
            1.,
            1.5,
            1.,
            1.,
            1.5,
            1.]
        self['Valve2022']['A_open'] = [
            0.00050078,
            0.0007021,
            0.00046806,
            0.00051776,
            0.00074755,
            0.00049837]
        self['Valve2022']['A_leak'] = [
            2.64705882e-04,
            2.64705882e-10,
            2.64705882e-10,
            2.64705882e-04,
            2.64705882e-10,
            2.64705882e-10]
        self['Valve2022']['l'] = 0.01626978
        self['Valve2022']['rho_b'] = 1050
        self['Valve2022']['papillary_muscles'] = False
        self['Valve2022']['papillary_muscles_slope'] = 100
        self['Valve2022']['papillary_muscles_min'] = 0.1
        self['Valve2022']['papillary_muscles_A_open_fac'] = 0.1
        self['Valve2022']['soft_closure'] = True

        self['Bag']['k'] = 10
        self['Bag']['V_ref'] = 0.0005
        self['Bag']['p_ref'] = 100

        self.set('Model.Peri.TriSeg.Y', 0.035)

        self.run(run_stable=True)

